@@ -110,6 +110,7 @@ enum GenericOptions: Int, CaseIterable, SectionType {
     case continueVideoPlayback
     case automaticallyPlayNextItem
     case enableTextScrollingInMediaList
+    case rememberPlayerState
 
     var description: String {
         switch self {
@@ -125,6 +126,8 @@ enum GenericOptions: Int, CaseIterable, SectionType {
             return "SETTINGS_NETWORK_PLAY_ALL"
         case .enableTextScrollingInMediaList:
             return "SETTINGS_ENABLE_MEDIA_CELL_TEXT_SCROLLING"
+        case .rememberPlayerState:
+            return "SETTINGS_REMEMBER_PLAYER_STATE"
         }
     }
 
@@ -142,6 +145,8 @@ enum GenericOptions: Int, CaseIterable, SectionType {
             return true
         case .enableTextScrollingInMediaList:
             return true
+        case .rememberPlayerState:
+            return true
         }
     }
 
@@ -159,6 +164,8 @@ enum GenericOptions: Int, CaseIterable, SectionType {
             return nil
         case .enableTextScrollingInMediaList:
             return nil
+        case .rememberPlayerState:
+            return nil
         }
     }
 
@@ -176,6 +183,8 @@ enum GenericOptions: Int, CaseIterable, SectionType {
             return kVLCAutomaticallyPlayNextItem
         case .enableTextScrollingInMediaList:
             return kVLCSettingEnableMediaCellTextScrolling
+        case .rememberPlayerState:
+            return kVLCPlayerShouldRememberState
         }
     }
 }
@@ -456,11 +465,14 @@ enum CastingOptions: Int, CaseIterable, SectionType {
 }
 
 enum AudioOptions: Int, CaseIterable, SectionType {
+    case preampLevel
     case timeStretchingAudio
     case audioPlaybackInBackground
 
     var description: String {
         switch self {
+        case .preampLevel:
+            return "SETTINGS_AUDIO_PREAMP_LEVEL"
         case .timeStretchingAudio:
             return "SETTINGS_TIME_STRETCH_AUDIO"
         case .audioPlaybackInBackground:
@@ -468,18 +480,29 @@ enum AudioOptions: Int, CaseIterable, SectionType {
         }
     }
 
-    var containsSwitch: Bool { return true }
+    var containsSwitch: Bool {
+        switch self {
+        case .preampLevel:
+            return false
+        default:
+            return true
+        }
+    }
     var subtitle: String? {
         switch self {
+        case .preampLevel:
+            return "6 dB"
         case .timeStretchingAudio:
             return "SETTINGS_TIME_STRETCH_AUDIO_LONG"
-        case .audioPlaybackInBackground:
+        default:
             return nil
         }
     }
 
     var preferenceKey: String? {
         switch self {
+        case .preampLevel:
+            return kVLCSettingDefaultPreampLevel
         case .timeStretchingAudio:
             return kVLCSettingStretchAudio
         case .audioPlaybackInBackground:
