@@ -26,6 +26,7 @@
 #import "VLCMediaFileDiscoverer.h"
 #import "VLCMigrationViewController.h"
 #import "VLCCloudStorageTableViewController.h"
+#import "VLCCloudStorageTableViewCell.h"
 #import "VLCMetadata.h"
 #import "VLCPlayerDisplayController.h"
 #import "NSString+SupportedMedia.h"
@@ -35,3 +36,5 @@
 #import "VLCStatusLabel.h"
 #import "VLCStoreController.h"
 #import "VLCThumbnailsCache.h"
+#import "VLCGoogleDriveTableViewController.h"
+#import "VLCGoogleDriveConstants.h"
